from .api_base import ApiBaseObject
import pandas

class Change(ApiBaseObject):

    calculated_fields = [
        "CI", "Category", "Compliant", "diff-action", "diff-object-name", "Dynamic", "is-ignore", 
        "Loneliness", "SystemTag", "actionString", "applicationLayer", "artifact", 
        "closed", "comment", "file-type", "freq", "impact-area", "it-action", 
        "logical-layer", 
        "relative-change", "server-component", "suspect", "time-anomaly", "type-violation", "is-toggle",
        "value-anomaly", "value-type", 
        "application-definition"
        ]

    _default_fields = ['AnnotatedDescription', 'Authorized', 'AuthorizedCause', 'CISubType', 'CIType', 'ChangeType', 'Class', 'Consistency', 'ConsistencyKey', 'ConsistencyPercent', 'CurrentValue', 'EnvClass', 'Environment', 'EnvironmentName', 'EnvironmentType', 'FullPath', 'History', 'Host', 'HostClass', 'HostType', 'ID', 'Icon', 'IsChange', 'Item', 'LastItem', 'LastPath', 'Name', 'NoiseCategory', 'Operation', 'Os', 'OsIcon', 'Parameter', 'Path', 'PreviousValue', 'RiskDescription', 'RiskLevel', 'RiskNodeId', 'RiskScenario', 'RiskSimpleLevel', 'Riskscore', 'Significances', 'StartTime', 'State', 'Time', 'Type', 'envId', 'hostId']
    
    def list(self, env_id, start=0, end="now", filter='', fields=None, details=True, count=100000, group_by=None, setup_analysis=False, **kwargs):
        """
        Get all changes for particular environment
        """
        if setup_analysis:
            self.set_up_analysis(env_id, start, end)

        if fields == None:
            fields = self.calculated_fields

        res = self.api.request("/html/scripts/changes-tree-summary.jsp", {
            "envId": env_id,
            "start": start,
            "end": end,
            "details": details,
            "count": count,
            "groupBy": group_by,
            "crit": filter,
            "field": ",".join(fields) if fields is not None else ""
            }, records_path="Next.Changes.Change", **kwargs)

        return self._normalize_fields(res, fields)


    def set_comment(self, change_ids, env_id, comment, **kwargs):
        """
        Add comments
        """
        if type(change_ids) is list:
            change_ids = ",".join(map(str, change_ids))

        
        # Only now you can Mark as suspect
        return self.api.request('/next/changes',{
            'action': 'comment',
            'ids': change_ids,
            'Comment': comment,
            'selectedTab': 'Analysis',
            'ChangesContext': 'Analysis',
            'returnStatus':False
            }, **kwargs)


    def set_suspect_tag(self, change_id, env_id, remove=None, start=0, end='now', **kwargs):
        """
        Add suspect tag to the specific change
        """
        action = 'tag' if remove == None else 'untag'

        self.set_up_analysis(env_id = env_id, start=start, end=end, **kwargs)
        
        # Only now you can Mark as suspect
        return self.api.request('/next/changes',{
            'action': action,
            'ids': change_id,
            'selectedTab': 'Analysis',
            'ChangesContext': 'Analysis',
            }, **kwargs)

    def set_authorization(self, change_id, env_id, state, start=0, end='now', **kwargs):
        """
        Set change as authorized
        """

        self.set_up_analysis(env_id = env_id, start=start, end=end, **kwargs)

        return self.api.request('/next/changes', {
            'action': 'setAuthorized',
            'ids':change_id,
            'cause': 'Authorization+changed+via+evolvenBot',
            'selectedTab': 'Analysis',
            'ChangesContext': 'Analysis',
            'state': state # 1: authorized, -1: unathorized, 0: no data
            }, **kwargs)


    def set_up_analysis(self, env_id, start=0, end='now', **kwargs):
        """
        Set up analysis properties.
        """
        # First, set analysis
        self.api.request('/next/analysisDashboard',{
            'action': 'getAnalysis',
            'envId': env_id,
            'start': start,
            'end': end,
            'reload': True,
            'monitoring': True,
            'filter': 0,
            }, **kwargs)

        # Set cluster
        self.api.request('/next/changes', {
            'action': 'setCluster',
            'ids': 0,
            'cluster': True,
            'envId': env_id,
            'start': start,
            'end': end,
            }, **kwargs)
        
        # Get all changes
        self.api.request('/next/changes', {
            'action': 'getChanges',
            'type': 'cluster',
            'ids': 0,
            'Monitoring': True,
            'envId': env_id,
            'start': start,
            'end': end,
            }, **kwargs)


    def fields(self):
        return self.api.request('/next/options', {"type":"diff-field"}, records_path="Next.Option")


    def _normalize_fields(self, res, extra_fields):

        if res is not None and type(res) == pandas.core.frame.DataFrame and len(res) > 0:

            if "AnnotatedDescription" in res:
                res["AnnotatedDescription"] = [
                    " ".join([el["Text"] for el in record]) 
                    for record in res["AnnotatedDescription"]]

            if "Operation" in res:
                operation_type = {
                    "130": "PROPERTY UPDATE", 
                    "131": "PROPERTY ADD",
                    "132": "PROPERTY REMOVE",
                    "133": "NODE ADD",
                    "134": "NODE REMOVE"
                }
                res["Operation"] = [operation_type[record] if record in operation_type else "?" 
                                        for record in res["Operation"]]

            # Extra fields
            cng_list = []
            for record in res.Fields:
                cng = {}
                for field in record.Field:        
                    cng[field.Name] = field.Value
                for field in extra_fields:
                    if field not in cng:
                        cng[field] = "null"
                cng_list.append(cng)

            cng_df = pandas.DataFrame.from_dict(cng_list)

            res = pandas.concat([res, cng_df], axis=1)
            res = res.drop(['Fields'], axis=1)

            # Make sure all the default fields are present
            for field in self._default_fields:
                if field not in res:
                    res[field] = [''] * len(res)

            # sort columns to make sure they appeary in the same order
            res = res.reindex_axis(sorted(res.columns), axis=1)

        return res

    def get_timeline(self, env_id, start=0, end='now', **kwargs):
        
        self.set_up_analysis(env_id=env_id, start=start, end=end)

        return self.api.request('/next/analysisDashboard', {
            'action': 'getTimeline',
            'type': 'cluster',
            'ids' : 0,
            'envId': env_id,
            'start': start,
            'end': end
            })





import pandas as pd
import json

class ApiResultObject(object):
    def __init__(self, dict_):
        self.__dict__.update(dict_)

    def __repr__(self):
        return "{%s}"%", ".join(["%s: %s"%(key, self.__dict__[key]) for key in self.__dict__])
    
    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True)
    
    def toDict(self):
        return self.__dict__

    def toDataFrame(self, records_path):
        object_path = records_path.split(".")
        records = self
        for obj in object_path:
            if obj in records.__dict__:
                records = getattr(records, obj)
            else:
                return pd.DataFrame()

        if records:
            records = records if type(records) == list else [records]
            if len(records) > 0:
                return pd.DataFrame.from_records([r.toDict() for r in records])
    
        return pd.DataFrame()


    # enables adding of attributes
    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]
    



from .api_base import ApiBaseObject


class Environment(ApiBaseObject):

    def search(self, term, env_id, **kwargs):
        return self._search_env(term, False, env_id)

    def search_logical_env(self, term, env_id=3, **kwargs):
        return self._search_env(term, False, env_id)

    def search_discovered_env(self, term, host_id=5, **kwargs):
        return self._search_env(term, False, host_id)

    def list(self, flat=True, **kwargs):
        res = self.api.request("/html/scripts/export-environments-tree.jsp?action=download", {
            "set-element-name": "true",
            "download": "false",
            "simple": "false",
            "stop-on-host": "false",
            "add-env-data": "true"
        })

        if not flat:
            return res

        res_list = []
        for root in res.Next.Root:
            if hasattr(root, "Host"):
                for host in root.Host:
                    self._flatten_env(host, res_list, "Environment")
            if hasattr(root, "Root"):
                for folder in root.Folder:
                    self._flatten_env(folder, res_list, "Folder")

        return res_list

    def get_children(self, env_id, **kwargs):
        return self.api.request("/next/Environments?action=children", {
            "envId": "%s" % env_id,
            "selectedTab": "Inventory",
            "ChangesContext": "Inventory"
        })

    def get_children_ids(self, env_id, **kwargs):
        return self.api.request("/next/Environments?action=descendants", {
            "EnvIdOnly": "true",
            "HasConfig": "true",
            "envId": "%s" % env_id,
            "selectedTab": "Inventory",
            "ChangesContext": "Analysis",
        })

    def _flatten_env(self, root, res_list, attr):
        if hasattr(root, attr):
            sub_env = getattr(root, attr)
            if type(sub_env) == list:
                for env in sub_env:
                    self._flatten_env(env, res_list, attr)
            else:
                self._flatten_env(sub_env, res_list, attr)
            delattr(root, attr)
        res_list.append(root)


    def _search_env(self, term, hosts_only=False, env_id=-1):
        params = {
            "Rows": -1,
            "Page": -1,
            "crit": term,
            "folder": "false",
            "idx": 0,
            "count": 0,
            "compareTo": 0,
            "filter": "false",
            "hosts":  hosts_only,
            "history": "false"}
        if int(env_id) > 0:
            params["env_id"] = env_id

        return self.api.request("/next/Environments?action=search", params, records_path="Next.Environment")

    def get_summary(self, env_id, **kwargs):
        return self.api.request("/next/Environments?action=summary", {
            "envId": "%s" % env_id
        }, **kwargs)



from .api_base import ApiBaseObject

class Blended(ApiBaseObject):

	def list(self, env_id, start=0, end='now', **kwargs):
		return self.api.request("/next/blended", {
			'action': 'get',
			'envId': env_id,
			'start': start,
			'end': end
			}, records_path="Next.Record", **kwargs)

	def get(self, blended_id, **kwargs):
		return self.api.request("/next/blended", {
			'action': 'get',
			'id': blended_id,
			}, records_path="Next.Record", **kwargs)


	def create(self, event_source, event_id, event_name, event_message, event_start, event_end, event_hosts, event_deeplink, **kwargs):
		return self.api.request("/next/blended", {
			'action': 'create',
			'delete': True,
			'Source': event_source,
			'eventId': event_id,
			'event': event_name,
			'Message': event_message,
			'Start': event_start,
			'End': event_end,
			'host': event_hosts,
			'DeepLink': event_deeplink,
			}, records_path="Next", **kwargs)


	def investigate(self, event_id, **kwargs):
		return self.api.request("/html/scripts/changes-tree-summary.jsp", {
			# 'GroupBy': 'analysis-filter',
			'blended-environment': event_id
			}, records_path="Next", **kwargs)



from .api_base import ApiBaseObject
import pandas

class Policy(ApiBaseObject):

    def list(self, **kwargs):
        """
        Get all policies
        """
        return self.api.request("/next/policy", {
            "action":"get",
            "returnStatus":"false",
            }, records_path="Next.Rule", **kwargs)


    def get(self, id, env_id=None, **kwargs):
        """
        Get policy details. If env_id is set, it returns policy results only for 
        specific environment (it could be logical env)
        """

        return self.api.request("/next/policy", {
            "action":"getResult",
            "json":"true",
            "RuleId":id,
            "envId":env_id,
            }, records_path="Next.Environment", **kwargs)







from .api_base import ApiBaseObject
import pandas as pd


class Host(ApiBaseObject):

    def __init__(self, api, discovered_root):
        self.api = api
        self.discovered_root = discovered_root
    
    def search(self, term, **kwargs):
        """
        Finds all the matches in hosts of discovered environments. Default installation holds discovered environments under root env_id=5.
        """
        return self.api.Environment._search_env(term, True, env_id=self.discovered_root)
    

    def get_by_name(self, name, **kwargs):
        """
        Returns the first host that matched given name
        """
        return self.search("^%s"%name)


    def get_by_id(self, term, **kwargs):
        """
        TODO: Returns host details by id.
        """
        return self.search("hostid=%s"%term)


    def list_discovered_env(self, detailed=True, stop_on_host=True, **kwargs):

        res = self.api.request("/html/scripts/export-environments-tree.jsp?action=download", {
            "set-element-name": "true",
            "download": "false",
            "simple": "false",
            "stop-on-host": stop_on_host,
            "add-env-data": detailed,
            "envId": self.discovered_root
            }, records_path="Next.Root.Host", **kwargs)

        return res

    def list(self, **kwargs):
        return self.api.request("/next/hosts?action=get", {}, records_path="Next.Host", **kwargs)




from .api_base import ApiBaseObject

class Login(ApiBaseObject):

	def login(self, user=None, password=None, **kwargs):
		res = self.api.request("/next/api?action=login", {
            'user': user if user else self.api.username,
            'pass': password if password else self.api.password
           })
		try:
			return res.Next.ID
		except:
			return None


from .api_base import ApiBaseObject


class AppDef(ApiBaseObject):
    
    def list(self, table="false", **kwargs):
        return self.api.request("/next/appDef?", {
            "action": "list",
            "table": table
        }, **kwargs)

    def get(self, guid, **kwargs):
        return self.api.request("/next/appDef?", {
            "action": "get",
            "guid": guid
            }, **kwargs)



from .api_base import ApiBaseObject

class EQL(ApiBaseObject):
    
    def execute(self, query, **kwargs):
        """
        Parses query and returns a deeplink to product.
        """
        if self.api.debug:
        	print(query)
        return self.api.request("/next/changes?action=getEvolvenQueryLanguage", {
            "content": query.replace(" ", "+")
            })




from future.standard_library import install_aliases
install_aliases()

import urllib.request, urllib.parse, urllib.error, urllib.request, urllib.error, urllib.parse

import json
import time
import random

from .res.change import Change
from .res.configuration import Configuration
from .res.environment import Environment
from .res.host import Host
from .res.eql import EQL
from .res.login import Login
from .res.appdef import AppDef
from .res.api_result import ApiResultObject
from .res.blended import Blended
from .res.policy import Policy



class EvolvenAPI():
    
    def __init__(self, host, port=443, username=None, password=None, 
                db_lib = "enlight_lib", db_main = "enlight_main", db_host=None,
                discovered_root_id=5, logical_root_id=3, debug=False, fake=False, 
                return_type="object", session_key=None, max_retries=5):

        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.db_lib = db_lib
        self.db_main = db_main
        self.db_host = db_host
        self.debug = debug
        self.fake = fake
        self.return_type = return_type
        self.max_retries = max_retries
        self.session_key_self_assigned = True


        self.Configuration  = Configuration(self)
        self.Change         = Change(self)
        self.Environment    = Environment(self)
        self.Host           = Host(self, discovered_root_id)
        self.EQL            = EQL(self)
        self.Login          = Login(self)
        self.AppDef         = AppDef(self)
        self.Blended        = Blended(self)
        self.Policy         = Policy(self)

        if session_key is None:
            s_key = self.Login.login()
            if s_key is not None:
                self.session_key = s_key
                self.session_key_self_assigned = False
            else:
                raise Exception("Cannot login")
        else:
            self.session_key = sessionKey
            self.session_key_self_assigned = False



    def reset_session_key(self):
        self.session_key = str(int(time.time()))+str(random.getrandbits(5))
        self.session_key_self_assigned = True

    def set_session_key(self):
        self.session_key = str(int(time.time()))+str(random.getrandbits(5))
        self.session_key_self_assigned = False


    def _get_api_url(self, action):

        if not "?" in action:
            action = action+"?"
           

        url = "%(host)s:%(port)s/enlight.server%(action)s&json=true&pretty=true"%{
            "host": self.host,
            "port": self.port,
            "action": action,
        }

        return url
        

    def request(self, action, parameters, records_path=None, return_type=None, return_values=True, **kwargs):
        
        url = self._get_api_url(action)
        
        # querystring = {
        #     "db_lib": self.db_lib,
        #     "db_main": self.db_main,
        #     "db_host": self.db_host
        # }

        querystring = {}
        if not self.session_key_self_assigned:
            querystring.update({
                "EvolvenSessionKey": self.session_key
                })
        else:
            querystring.update({
                "user": self.username,
                "pass": self.password,
                }) 

        querystring.update(parameters)
        querystring.update(kwargs)

        url += "&" + urllib.parse.urlencode(querystring)


        if self.debug:
            print(url)

        if not self.fake:
            res = None
            n_retries = 0

            while res is None and n_retries < self.max_retries:
                try:
                    req = urllib.request.urlopen(url, timeout=120)
                    res = req.read().decode('ascii', 'ignore')
                except urllib.error.URLError:
                    n_retries +=1
                    print('URL request failed (%s/%s). Trying again.'%(n_retries, self.max_retries))
                    time.sleep(2)
                    pass

            if return_values:
                return self._parse_response(res, url, records_path, return_type)
        
        return "{}"



    def _dict2obj(self, d):
        return json.loads(json.dumps(d), object_hook=ApiResultObject)
    
    def _parse_response(self, response, url=None, records_path=None, return_type=None):

        if return_type is None:
            return_type = self.return_type
        
        data = json.loads(response)
                
        if url is not None and type(data) is not list:
            data["url"] = url

        # return as string
        if return_type == "string":
           return json.dumps(data)

        # return as json
        elif return_type == "json":
            return data

        # return as DataFrame
        elif return_type == "DataFrame":
            if records_path is not None:
                res = self._dict2obj(data)
                return res.toDataFrame(records_path)

        # return as obj by default
        return self._dict2obj(data)


    def search_to_count(self, result):
        return int(result["Next"]["Count"])




from .api_base import ApiBaseObject


class Configuration(ApiBaseObject):

	ELEMENT_TYPE = {
		"34301": "Environment",
		"34302": "Tree",
		"34303": "Node",
		"34304": "Attribute",
		"34305": "Table",
		"34306": "Table Row",
		}
	
	def search(self, term, env_id, search_children=True, search_tables=True, **kwargs):
		"""
		Finds all the CIs under specific environment and its' children.
		"""
		# get all childs
		childs = []

		if search_children:
			data = self.api.Environment.get_children_ids(env_id).Next.envId
			childs = data if type(data) == list else []
			childs.append("%s"%env_id)

		return self.api.request("/next/configurationTree", {
			"action": "search",
			"Rows":"-1",
			"Page":"1",
			"Search":"%s"%term,
			"envId":"%s"%env_id,
			"Att":"true",
			"Table":"%s"%search_tables,
			"offset":"0",
			"selectedTab":"Inventory",
			"ChangesContext":"Dashboard",
			"term": term,
			"env_id": ",".join(childs),
			"search_tables": search_tables,
			}, records_path="Next.Node", **kwargs)


	def list(self, env_id, **kwargs):
		return self.api.request("/next/configurationTree", {
			"action": "setEnv",
			"envId": env_id
		}, records_path="Next.Node", **kwargs)



	def get_children(self, node_id, env_id=None, **kwargs):
		self.__init_config_tree(env_id)
		return self.api.request("/next/configurationTree", {
			"action": "children",
			"nodeId":"%s"%node_id,
			"Rows":"-1",
			"Page":"1"
			}, records_path="Next.Node", **kwargs)


	def get_node_by_path(self, env_id, node_path, parent_node=None):

		if not type(node_path) is list:
			node_path = node_path.split("/")
		
		nodes = None
		if parent_node is not None:
			nodes = self.get_children(node_id=parent_node.ID)
		else:
			nodes = self.list(env_id=env_id)
			
		for idx, node in nodes.iterrows():
			if node.Name == node_path[0]:
				if len(node_path) == 1:
					return node
				return self.get_node_by_path(env_id, node_path[1:], node)
		
		return None


	# def get(self, node_id, env_id=None, **kwargs):
	# 	self.__init_config_tree(env_id)
	# 	return self.api.request("/next/configurationTree", {
	# 		"action": "details",
	# 		"id": "%s"%node_id
	# 		}, **kwargs)

	# def get_table(self, node_id, env_id=None, **kwargs):
	# 	self.__init_config_tree(env_id)
	# 	return self.api.request("/next/configurationTree", {
	# 		"action": "table",
	# 		"nodeId": "%s"%node_id,
	# 		"selectedTab": "Inventory",
	# 		"ChangesContext": "Dashboard",
	# 		}, **kwargs)


	def __init_config_tree(self, env_id):
		if env_id is not None:
			self.list(env_id)




from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='evolven',
    version='1.0.0',

    description='Evolven API Python Client',
    long_description_content_type='text/markdown',
    long_description=long_description,

    url='https://github.com/evolven-software/evolven-api-python',

    author='Evolven Data Science Team',
    author_email='bostjan@evolven.com',

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Scientific/Engineering :: Information Analysis',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],

    keywords='evolven api development',

    packages=find_packages(exclude=['examples']),
)